""" Copyright start
  Copyright (C) 2008 - 2022 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import json
import pika
from datetime import datetime

from helpers.es_helper import publish_options, index_options, data_options
from util.logger import get_logger
from elasticsearch.exceptions import RequestError, NotFoundError, AuthenticationException

logger = get_logger('__name__')


class Listener(object):
    def __init__(self, channel):
        self.channel = channel

    def listen(self):
        self.channel.basic_consume(queue='queue.falcon.data', on_message_callback=self.on_data_message)
        self.channel.basic_consume(queue='queue.falcon.search', on_message_callback=self.on_data_search)
        self.channel.basic_consume(queue='queue.falcon.system', on_message_callback=self.on_publish_message)
        self.channel.basic_consume(queue='queue.falcon.status', on_message_callback=self.on_status_message)
        try:
            self.channel.start_consuming()
        except KeyboardInterrupt:
            self.channel.stop_consuming()

    def on_data_message(self, channel, method, properties, body):
        """Called when a data publish message is received.
        This will be called on CUD operation on any record.
        It is sent by CRUD-HUB. body should contain two mandatory keys 'esoperation' and 'data'. Log message and ack it."""
        try:
            request_data = json.loads(bytes.decode(body))
            for data in request_data:
                if 'esoperation' in data and 'data' in data and data_options[data['esoperation']]:
                    try:
                        change_data = data['data'].get('changeData') if data['data'].get('changeData') else data[
                            'data'].get(data_options[data['esoperation']])
                        index_options[data['esoperation']](data['data'].get(data_options[data['esoperation']]),
                                                           change_data)
                    except Exception as e:
                        logger.exception("Error Occurred While Inserting Data: '%s'" % str(e))
                else:
                    pass
        except Exception as e:
            logger.exception('Invalid json message. Cannot index this data')
        self.channel.basic_ack(delivery_tag=method.delivery_tag)

    def _handle_response(self, props, method, data):
        self.channel.basic_publish(exchange='', routing_key=props.reply_to,
                                   properties=pika.BasicProperties(correlation_id=props.correlation_id),
                                   body=json.dumps(data))
        self.channel.basic_ack(delivery_tag=method.delivery_tag)

    def on_data_search(self, channel, method, properties, body):
        """Called when a data search message is received.
        This will be called on Global Search of records
        It is sent by CRUD-HUB. body should contain two mandatory keys 'esoperation' and 'payload'. Log message and ack it."""
        request_data = json.loads(bytes.decode(body))
        if 'esoperation' in request_data[0] and 'payload' in request_data[0]:
            try:
                logger.debug("Serving Falcon Search Request at %s" % str(datetime.now()))
                search_data = index_options[request_data[0]['esoperation']](request_data[0]['payload'])
            except RequestError as e:
                logger.exception(
                    "Failed to Search Data Exception occurred for input: '%s'" % (str(request_data[0]['payload'])))
                search_data = {"error": str(e), 'status_code': 400}
            except NotFoundError as e:
                logger.exception(
                    "Failed to Search Data Exception occurred for input: '%s'" % (str(request_data[0]['payload'])))
                search_data = {"error": str(e), 'status_code': 404}
            except AuthenticationException  as e:
                logger.exception(
                    "Failed to Search Data Exception occurred for input: '%s'" % (str(request_data[0]['payload'])))
                search_data = {"error": str(e), 'status_code': 401}
            except Exception as e:
                logger.exception(
                    "Failed to Search Data Exception occurred for input: '%s'" % (str(request_data[0]['payload'])))
                search_data = {"error": str(e), 'status_code': 500}

            self._handle_response(properties, method, search_data)


    def on_status_message(self, channel, method, properties, body):
        """Called when a status message is received.
        It is sent by CRUD-HUB. Log message and ack it."""
        request_data = json.loads(bytes.decode(body))
        if 'esoperation' in request_data[0]:
            try:
                logger.debug("Serving Falcon Search Request at %s" % str(datetime.now()))
                search_data = index_options[request_data[0]['esoperation']](request_data[0]['payload'])
            except Exception as e:
                logger.exception(
                    "Fail to Search Data Exception occurred for input: '%s'" % (str(request_data[0]['payload'])))
                search_data = {"error": str(e)}
            self._handle_response(properties, method, search_data)


    def on_publish_message(self, channel, method, properties, body):
        """
        Called when a system publish message is received.
        This will be called on system publish esoperation. It is sent by CRUD-HUB.
        Body should contain two mandatory keys 'esoperation' and 'payload'. Log message and ack it.
        """
        logger.info('on_publish_message called')
        payload_data = json.loads(bytes.decode(body))
        try:
            publish_options[payload_data['esoperation']](payload_data['payload'], channel.connection)
            data = {"success": True}
        except Exception as e:
            data = {"success": False}
            logger.exception("Fail to Publish for Operation: '%s'" % (payload_data['esoperation']))
        if properties.reply_to:
            self._handle_response(properties, method, data)
        else:
            self.channel.basic_ack(delivery_tag=method.delivery_tag)
