""" Copyright start
  Copyright (C) 2008 - 2022 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import settings
from constants import (FRAGMENT_SIZE, NUMBER_OF_FRAGMENT,
                       PREFIX_INDEX, SOURCE_FIELDS, CONTEXT_CODE)
from elasticsearch_dsl import Search, Q
from util.connection import es
from util.logger import get_logger

logger = get_logger('__name__')


class SearchElastic():

    def __init__(self, searchObj):
        self.index = searchObj.get('index')
        self.filter_term = searchObj.get('filter_term')
        self.user_filter = [searchObj.get('current_user')]
        self.search_term = searchObj.get('search_term')
        # If sort order is not passed it will by default solve by relevance
        self.sort = searchObj.get('sort', '_score')
        self.offset = searchObj.get('offset', 0)
        self.size = searchObj.get('size', 30)
        self.from_date = searchObj.get('recordModifiedFromDt', 0)
        self.to_date = searchObj.get('recordModifiedToDt', 0)
        self.search_type = searchObj.get('searchType', '_default')

    def prepare_list_index(self):
        index_list = [PREFIX_INDEX + item for item in self.index if item != 'files']
        if 'files' in self.index:
            index_list.append('files_prod')
        available_index = sorted(es.indices.get_alias("*").keys())
        return [index for index in index_list if index in available_index]

    def prepare_search_term(self):
        if self.search_type == 'phrase_prefix':
            return [Q("multi_match", query=item, fields=[],type='phrase_prefix') for item in self.search_term]

        return [Q("multi_match", query=item, fields=[]) for item in self.search_term]

    def prepare_filter_term(self):
        first_filter = \
            (Q("terms", **{"owners": self.filter_term}) & Q("exists", field='owners')) | \
            (Q("terms", **{"userOwners": self.user_filter}) & (Q("exists", field='userOwners')) |
             (~Q("exists", field='owners')))
        return first_filter

    def prepare_query(self, index_list, must_match, filter_match):
        dsl_search = Search(using=es, index=index_list)
        return dsl_search.query(Q('bool', must=must_match, filter=filter_match)). \
                   highlight('*', fragment_size=FRAGMENT_SIZE, number_of_fragments=NUMBER_OF_FRAGMENT). \
                   sort(self.sort)[self.offset:self.size]

    def search(self):
        '''
        must_match: Will contribute to score and will use AND operator between multiple queries
        should_match: Will not contribute to score and will use OR operator between multiple queries
        sort: Will sort the result as per relevance by default, user can specify specific param too
         pagination: Controlled by python slicing api by default 30 record will return
        '''
        logger.info("Search request received")
        index_list = self.prepare_list_index()
        must_match = self.prepare_search_term()
        filter_match = self.prepare_filter_term()
        search_result = self.prepare_query(index_list, must_match, filter_match)
        if self.from_date and self.to_date:
            search_result = search_result.filter(
                Q("range", **{"modifyDate": {'gte': self.from_date, 'lte': self.to_date}}))
        elif self.from_date:
            search_result = search_result.filter(Q("range", **{"modifyDate": {'gte': self.from_date}}))
        elif self.to_date:
            search_result = search_result.filter(Q("range", **{"modifyDate": {'lte': self.to_date}}))
        search_result = search_result.extra(_source={'include': SOURCE_FIELDS})
        return {
            "search_result": search_result.execute().to_dict().get('hits'),
            "context_code": CONTEXT_CODE.get(settings.is_published)
        }


class SearchElasticRaw():
    '''
    Will be used if Client wants to perform Raw Query Search Over Elastic Search through Falcon
    Currently will not be in used
    '''
    def __init__(self, searchObj):
        self.raw_query = searchObj.get('raw_query')
        self.offset = searchObj.get('offset', 0)
        self.size = searchObj.get('size', 30)

    def search(self):
        dsl_search = Search(using=es, index=[])
        search_result = dsl_search.from_dict(self.raw_query)[self.offset:self.size]
        return search_result.execute().to_dict().get('hits')
