""" Copyright start
  Copyright (C) 2008 - 2022 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import json
import math
import re
from copy import deepcopy

from elasticsearch import helpers

import core.prediction as prediction
import settings
from constants import (CH_TO_ES_MAPPING, BATCH_SIZE, PREFIX_INDEX,
                       API_PREFIX, FILE_INDEX, DISPLAY_NAME_REGEX,
                       WORKFLOW_INDEX, RULE_INDEX, IGNORE_MODULES,
                       BUILTIN_MODULE, WORKFLOW_FIELDS, RULE_FIELDS,
                       MAPPING_SETTINGS, WORKFLOW_SETTINGS, RULE_SETTINGS, FILES_SETTINGS)
from core.search import SearchElastic
from helpers.helper import get_mmd_meta
from util.communicator import _make_request
from util.config import config
from util.connection import es
from util.logger import get_logger

logger = get_logger('__name__')


def delete_tenant_data(tenant_iri):
    if settings.mmd:
        mmd = settings.mmd
    else:
        try:
            settings.mmd = get_mmd_meta()
            mmd = settings.mmd
        except Exception as e:
            logger.exception(str(e))
            mmd = {}
    for item in mmd.keys():
        if mmd.get(item, {}).get('peer_replicable', False):
            es.delete_by_query(index=PREFIX_INDEX + item,
                               ignore=[400, 404],
                               body={'query': {'term': {'tenant.keyword': tenant_iri}}})


def record_index_delete(data, change_data=None):
    if data.get('@id'):
        logger.info("Request for deleting data, %s", data['@id'])
        module_type = data['@id'].split('/')[3]
        if module_type == 'tenants':
            delete_tenant_data(data['@id'])
        if es.indices.exists(index=PREFIX_INDEX + module_type):
            for key, value in data.items():
                if type(value) == dict and '@type' in value and value['@type'] == 'File' and \
                        value['@id'].split('/')[3] == 'files':
                    pass
                elif type(value) == list and config.get('enable_reverse_mapping', fallback=True):
                    reverse_data_delete(data['@id'], key, value)
            es.delete(index=PREFIX_INDEX + module_type,
                      id=data['@id'].split('/')[4],
                      ignore=[400, 404])


def record_index_bulk_delete(data, change_data=None):
    uuids = [item.get('uuid') for item in data]
    query = {"query": {"term": {"uuid.keyword": uuids}}}
    es.delete_by_query(index=PREFIX_INDEX + change_data,
                       body=query,
                       ignore=[400, 404])


def record_index_search(data):
    search_obj = SearchElastic(data)
    return search_obj.search()


def record_index_status(data):
    index_status = []
    all_index = es.cat.indices(format='json')
    for index_metadata in all_index:
        if index_metadata.get('index').find('cyops_', 0) >= 0:
            index_status.append(
                {'index_name': index_metadata.get('index')[6:-1], 'index_count': index_metadata.get('docs.count', 0),
                 'index_health': index_metadata.get('health', 'red')})
        elif index_metadata.get('index') == 'files_prod':
            index_status.append({'index_name': 'files', 'index_count': index_metadata.get('docs.count', 0),
                                 'index_health': index_metadata.get('health', 'red')})

    return index_status


def get_mmd():
    if settings.mmd:
        mmd = settings.mmd
    else:
        try:
            settings.mmd = get_mmd_meta()
            mmd = settings.mmd
        except Exception as e:
            logger.exception(str(e))
            mmd = {}
    return mmd


def filter_data(module_type, bulk, copy_data, data, change_data, indexed_data=dict()):
    mmd = get_mmd()
    regex_pattern = re.compile(DISPLAY_NAME_REGEX)

    exclude_list = mmd.get(module_type, {}).get('encrypted_fields', [])
    datetime_fields = mmd.get(module_type, {}).get('datetime_fields', [])
    display_name = mmd.get(module_type, {}).get('display_name', '{{ name }}')
    json_fields = mmd.get(module_type, {}).get('json_field', [])
    file_fields = mmd.get(module_type, {}).get('file_field', [])
    ignore_reverse_index_module = mmd.get(
        module_type, {}).get('ignore_reverse_index_module', [])

    for ele in regex_pattern.finditer(display_name):
        display_name = display_name.replace(
            ele.group(), str(data.get(ele.groups()[0].strip(), '')))
    for key, value in copy_data.items():
        if key in exclude_list:
            del data[key]
        elif key in datetime_fields and value:
            data[key] = value * 1000
        elif key in json_fields:
            if value:
                data[key] = json.dumps(value)
            else:
                del data[key]
        elif type(value) == list:
            if key == 'recordTags':
                data['recordTags'] = value
            if key in ignore_reverse_index_module:
                continue
            if config.get('enable_reverse_mapping', fallback=True) and not bulk:
                reverse_data_relate(
                    data['@id'], key, value, change_data, indexed_data)
        elif key in file_fields and value:
            continue
        if type(value) == dict and key not in json_fields:
            data[key] = value.get('@id', '')
    data['displayName'] = display_name
    data['uuid'] = data['@id'].split('/')[4]
    if data.get('@type', "") == 'Comment':
        del (data['content'])
    return data


def filter_builtin_module_data(module_type, data):
    if module_type in 'workflows':
        return {k: data.get(k) for k in WORKFLOW_FIELDS if data.get(k)}
    elif module_type in 'rules':
        return {k: data.get(k) for k in RULE_FIELDS if data.get(k)}
    elif module_type in 'workflow_collections':
        if data.get('workflows'):
            for workflow in data.get('workflows'):
                record_index_data(workflow)
            del (data['workflows'])
    return data


def record_index_data(data, change_data=None, bulk=False):
    if data.get('@id'):
        logger.info("Request for indexing data, %s", data['@id'])
        module_type = data['@id'].split('/')[3]
        if module_type in BUILTIN_MODULE:
            data = filter_builtin_module_data(module_type, data)
        if es.indices.exists(index=PREFIX_INDEX + module_type):
            copy_data = deepcopy(data)
            if '@id' in data and module_type not in IGNORE_MODULES:
                data = filter_data(module_type, bulk,
                                   copy_data, data, change_data)
                if not bulk:
                    es.index(index=PREFIX_INDEX + module_type,
                             id=data['@id'].split('/')[4], body=data)
                else:
                    return data


def record_index_update_data(data, change_data=None, bulk=False):
    if data.get('@id'):
        logger.info("Request for updating data, %s", data['@id'])
        module_type = data['@id'].split('/')[3]
        if module_type in BUILTIN_MODULE:
            data = filter_builtin_module_data(module_type, data)
        if es.indices.exists(index=PREFIX_INDEX + module_type):
            copy_data = deepcopy(data)
            if '@id' in data and module_type not in IGNORE_MODULES:
                existing_query = es.get(
                    index=PREFIX_INDEX + module_type, id=data['@id'].split('/')[4], ignore=[400, 404])
                if type(existing_query) == dict:
                    existing_item = existing_query.get('_source')
                    if existing_item:
                        data = filter_data(
                            module_type, bulk, copy_data, data, change_data, existing_item)
                        existing_item.update(data)
                        if not bulk:
                            es.index(index=PREFIX_INDEX + module_type,
                                     id=data['@id'].split('/')[4], body=existing_item)
                        else:
                            return data
                    else:
                        data = filter_data(
                            module_type, bulk, copy_data, data, change_data)
                        es.index(index=PREFIX_INDEX + module_type,
                                 id=data['@id'].split('/')[4], body=data)


def reverse_data_relate(primary_record_uuid, key, value, change_data, indexed_data):
    """
    primary_record_uuid: Record for which the change has been trigger
    key: Name of Related record for which the reverse side is being modified
    value: Current Value of the related record for which the alerts need to be added
    changed_data: list of values that are added or removed in user requests
    indexed_data: existing data for same primary record
    """
    module_type = primary_record_uuid.split('/')[3]
    if key == module_type and config.get('disable_self_linking', fallback=True): return

    reverse_exclude_list = config.get('reverse_exclude_list') if config.get('reverse_exclude_list') \
        else ['people', 'owners', 'recordTags', 'workflows', 'rules', 'steps', 'routes', 'userOwners']
    reverse_exclude_list.append('recordTypes')
    reverse_exclude_list.append('recordTypeRules')
    if key not in reverse_exclude_list and change_data and key in change_data.keys():
        related_iri = [item.split('/')[4]
                       for item in (set(value + indexed_data.get(key, [])))]
        existing_item = es.mget(index=PREFIX_INDEX + key,
                                body={'ids': related_iri}, ignore=[400, 404])
        action = []
        for item in existing_item.get('docs', []):
            existing_data = item.get('_source')
            if existing_data and module_type in existing_data.keys():
                # Check If Value exist in latest record
                if existing_data.get('@id') in value:
                    if primary_record_uuid not in existing_data[module_type]:
                        existing_data[module_type].append(primary_record_uuid)
                # If Value is not existing in values it means, the value is removed
                # Check if it exists and if exist remove from the record
                elif primary_record_uuid in existing_data[module_type]:
                    existing_data[module_type].remove(primary_record_uuid)
                action.append({
                    '_index': PREFIX_INDEX + key,
                    '_id': existing_data['@id'].split('/')[4],
                    '_source': existing_data
                })

        if action:
            helpers.bulk(es,
                         action,
                         initial_backoff=config.get(
                             'elasticsearch', 'initial_backoff'),
                         max_backoff=config.get(
                             'elasticsearch', 'max_backoff'),
                         raise_on_error=False, stats_only=True,
                         request_timeout=90)


def reverse_data_delete(primary_record_uuid, key, value):
    module_type = primary_record_uuid.split('/')[3]
    if key == module_type and config.get('disable_self_linking', fallback=True): return
    if key not in config.get('reverse_exclude_list') if config.get('reverse_exclude_list') \
            else ['people', 'owners', 'recordTags', 'workflows', 'rules', 'steps', 'routes', 'userOwners']:
        related_iri = [item.split('/')[4] for item in value]
        existing_item = es.mget(index=PREFIX_INDEX + key,
                                body={'ids': related_iri}, ignore=[400, 404])
        action = []
        for item in existing_item.get('docs', []):
            existing_data = item.get('_source')
            if existing_data and module_type in existing_data.keys():
                if primary_record_uuid in existing_data[module_type]:
                    existing_data[module_type].remove(primary_record_uuid)
                action.append({
                    '_index': PREFIX_INDEX + key,
                    '_id': existing_data['@id'].split('/')[4],
                    '_source': existing_data
                })
        if action:
            helpers.bulk(es, action,
                         initial_backoff=config.get(
                             'elasticsearch', 'initial_backoff'),
                         max_backoff=config.get(
                             'elasticsearch', 'max_backoff'),
                         raise_on_error=False, stats_only=True,
                         request_timeout=90)


def publish_modify_index(data, rabbitmq_connection):
    existing_mapping = es.indices.get_mapping()
    existing_model = existing_mapping.keys()
    logger.debug('Data Received For Updating The Index : {0}'.format(data['type']))
    for key in data['type']:
        data_value = data['type'][key]
        index_name = PREFIX_INDEX + key
        if index_name in existing_model:
            existing_fields = existing_mapping[index_name]['mappings']['properties'].keys()
            mapping_data = get_modified_mapping(data_value, existing_fields)
            if mapping_data:
                properties = {}
                properties['properties'] = mapping_data
                mapping_response = es.indices.put_mapping(index=index_name, body=properties, ignore_unavailable=True,
                                                          master_timeout='240s')

                if not mapping_response.get('acknowledged', False):
                    logger.warn(
                        'Index Modification Request Failed: "%s" Unsuccessful.' % (properties))


def create_field_mapping(data, mapping_data={}):
    for attribute in data['attributes']:
        if attribute['formType'] in CH_TO_ES_MAPPING:
            mapping_data[attribute['name']] = CH_TO_ES_MAPPING[attribute['formType']]

    if data.get('trackable'):
        mapping_data['modifyDate'] = CH_TO_ES_MAPPING['datetime']
        mapping_data['createDate'] = CH_TO_ES_MAPPING['datetime']
        mapping_data['createUser'] = CH_TO_ES_MAPPING['picklist']
        mapping_data['modifyUser'] = CH_TO_ES_MAPPING['picklist']

    if data.get('softDeleteable', False):
        mapping_data['deletedAt'] = CH_TO_ES_MAPPING['datetime']

    if data.get('peerReplicable'):
        mapping_data['tenant'] = CH_TO_ES_MAPPING['picklist']

    if data.get('taggable'):
        mapping_data['recordTags'] = CH_TO_ES_MAPPING['manyToMany']

    if data.get('ownable'):
        mapping_data['owners'] = CH_TO_ES_MAPPING['manyToMany']

    if data.get('userOwnable'):
        mapping_data['userOwners'] = CH_TO_ES_MAPPING['manyToMany']

    if data.get('type') == 'comments':
        mapping_data['displayName'] = CH_TO_ES_MAPPING['text']
    else:
        mapping_data['displayName'] = {'type': 'text', 'analyzer': 'ngram_analyzer'}
    mapping_data['@id'] = CH_TO_ES_MAPPING['keyword']
    mapping_data['@type'] = CH_TO_ES_MAPPING['keyword']
    mapping_data['uuid'] = CH_TO_ES_MAPPING['keyword']
    mapping_data['id'] = {'type': 'integer'}
    return mapping_data


def get_modified_mapping(model_data, existing_fields):
    mapping_data = {}
    for key in model_data.get('added', {}).get('attributes', {}):
        if model_data['added']['attributes'][key]['formType'] in CH_TO_ES_MAPPING and key not in existing_fields and \
                model_data['added']['attributes'][key]['formType'] != 'password' and not \
                model_data['added']['attributes'][key]['encrypted']:
            mapping_data[key] = CH_TO_ES_MAPPING[model_data['added']['attributes'][key]['formType']]

    if model_data.get('modified', {}).get(
            'trackable') and 'createUser' not in existing_fields and 'modifyDate' not in existing_fields:
        mapping_data['modifyDate'] = CH_TO_ES_MAPPING['datetime']
        mapping_data['createDate'] = CH_TO_ES_MAPPING['datetime']
        mapping_data['createUser'] = CH_TO_ES_MAPPING['picklist']
        mapping_data['modifyUser'] = CH_TO_ES_MAPPING['picklist']

    if model_data.get('modified', {}).get('peerReplicable') and 'tenant' not in existing_fields:
        mapping_data['tenant'] = CH_TO_ES_MAPPING['picklist']

    if model_data.get('modified', {}).get('taggable') and 'recordTags' not in existing_fields:
        mapping_data['recordTags'] = CH_TO_ES_MAPPING['manyToMany']

    if model_data.get('modified', {}).get('ownable') and 'owners' not in existing_fields:
        mapping_data['owners'] = CH_TO_ES_MAPPING['manyToMany']

    if model_data.get('modified', {}).get('userOwnable') and 'userOwners' not in existing_fields:
        mapping_data['userOwners'] = CH_TO_ES_MAPPING['manyToMany']

    return mapping_data


def identify_required_changes(model_list):
    model_to_create = []
    existing_mapping = es.indices.get_mapping()
    existing_model = existing_mapping.keys()
    module = model_list['modelToCreate'][0]
    logger.info('Data Recived For re indexing : {0}'.format(module.get('type')))
    # receiving meta information along with model data from CH
    found_diff = False
    if module.get('meta') and len(module.get('meta')) > 0:
        logger.debug('Meta Data Recived: {0}'.format(module.get('meta')))
        meta_info = module.pop('meta')
        index_name = PREFIX_INDEX + module['type']
        if index_name in existing_model and meta_info.get('modified', {}).get('attributes', {}):
            new_mapping = create_field_mapping(meta_info.get('modified'))
            existing_fields = existing_mapping[index_name]['mappings']['properties'].keys()
            for index_map in new_mapping.keys():
                if index_map in existing_fields:
                    if existing_mapping[index_name]['mappings']['properties'][index_map] != new_mapping[index_map]:
                        found_diff = True
                        break
    if found_diff:
        # done to preserve old data format
        model_to_create.append(module)
        return model_to_create

    return model_list['modelToCreate']


def publish_create_index(models, rabbitmq_connection):
    """Create a mapping of the requested model in Elasticsearch. Log message and ack it."""
    settings.mmd = get_mmd_meta()
    populate_type = {}
    populate_type['modelToCreate'] = []
    if len(models['modelToCreate']):
        model_to_create = identify_required_changes(models)
    else:
        model_to_create = models
    if len(settings.mmd):
        for data in model_to_create:
            mapping_data = create_field_mapping(data)
            setting = {}
            properties = {}

            properties['properties'] = mapping_data
            setting['mappings'] = properties
            setting['settings'] = MAPPING_SETTINGS
            logger.info('creating index with mapping for module %s', str(data['type']))
            if es.indices.exists(index=PREFIX_INDEX + data['type']):
                es.indices.delete(index=PREFIX_INDEX + data['type'], ignore=[400, 404])
            es.indices.create(index=PREFIX_INDEX + data['type'], body=setting)
            populate_type['modelToCreate'].append(data['type'])

        if not es.indices.exists(index=FILE_INDEX):
            es.indices.create(index=FILE_INDEX, ignore=[400, 404], body=FILES_SETTINGS)
        if not es.indices.exists(index=WORKFLOW_INDEX):
            es.indices.create(index=WORKFLOW_INDEX, ignore=[400, 404], body=WORKFLOW_SETTINGS)
            populate_type['modelToCreate'].append('workflows')
        if not es.indices.exists(index=RULE_INDEX):
            es.indices.create(index=RULE_INDEX, ignore=[400, 404], body=RULE_SETTINGS)
            populate_type['modelToCreate'].append('rules')

        publish_populate_index(populate_type, rabbitmq_connection)


def publish_delete_index(data, rabbitmq_connection):
    for indexName in data:
        es.indices.delete(index=indexName, ignore=[400, 404])


def publish_populate_index(data, rabbitmq_connection):
    """Populate the Elasticsearch data.
    Log message and ack it."""
    settings.mmd = get_mmd_meta()
    settings.is_published = 2
    data['modelToCreate'].remove(
        'people') if 'people' in data['modelToCreate'] else data['modelToCreate']
    logger.info('Module Currently Getting Published: ' +
                str(data['modelToCreate']) + '\n')

    for model_name in data['modelToCreate']:
        logger.info('Populating index for module %s', str(model_name))
        model_base_url = 'https://' + config.get('crudhub', 'host') + \
                         API_PREFIX + model_name + \
                         '?$asJson=true&$limit=' + str(BATCH_SIZE)
        if model_name not in BUILTIN_MODULE:
            model_base_url += '&$relationships=true'

        headers = {'Content-Type': 'application/json'}
        try:
            api_data = _make_request(
                'get', model_base_url, body=None, headers=headers, allow_503=False)
        except Exception as e:
            logger.exception(
                'Failed to insert Bulk data for the url : ' + model_base_url)
            continue

        total_record = api_data.get('hydra:totalItems', 0)
        logger.info(
            'Indexing for Module: "%s" started Total Records to be indexed: "%s" \n' % (model_name, total_record))
        page_count = math.ceil(total_record / BATCH_SIZE)
        if total_record:
            while page_count > 0:
                url = model_base_url + '&$page=' + str(page_count)
                try:
                    api_data = _make_request('get', url, body=None, headers=headers)
                    es_index_data(model_name, api_data)
                except Exception as e:
                    settings.is_published = 3
                    logger.info('Publish Module: "%s" Unsuccessful. Total Records: "%s" Remaining Records: "%s"' %
                                (model_name, total_record, str(page_count * BATCH_SIZE)))
                    logger.exception('Failed to insert Bulk data for the url : ' + url)
                page_count -= 1
                rabbitmq_connection.process_data_events()

        logger.info("Module: '%s' Successful Total Records indexed: '%s'" % (model_name, total_record))
    settings.is_published = 1


def es_index_data(model_name, api_data):
    action = []
    if es.indices.exists(index=PREFIX_INDEX + model_name):
        for data_member in api_data['hydra:member']:
            processed_data = record_index_data(data_member, bulk=True)
            if processed_data:
                action.append({
                    '_index': PREFIX_INDEX + model_name,
                    '_id': processed_data['@id'].split('/')[4],
                    '_source': processed_data
                })

        helpers.bulk(es,
                     action,
                     initial_backoff=config.get('elasticsearch', 'initial_backoff'),
                     max_backoff=config.get('elasticsearch', 'max_backoff'),
                     raise_on_error=False, stats_only=True,
                     request_timeout=90)


def score_based_aggregation(results, aggs_fields):
    aggs = {}
    for record in results:
        for aggs_field in aggs_fields:
            if aggs_field not in aggs:
                aggs[aggs_field] = {}
            field_value = record['_source'].get(aggs_field)
            if not field_value:
                continue
            if field_value not in aggs[aggs_field]:
                aggs[aggs_field][field_value] = record['_score']
            else:
                aggs[aggs_field][field_value] = aggs[aggs_field][field_value] + \
                                                record['_score']
    buckets = {}
    for aggs_field in aggs_fields:
        buckets[aggs_field] = [{'key': field_value, 'score': score} for (
            field_value, score) in sorted(aggs[aggs_field].items(), key=lambda x: x[1], reverse=True)]
    return buckets


def record_index_similar(payload):
    index = payload['index']
    ids = payload.pop('id', [])
    if not ids: ids = []
    docs = payload.pop('docs', [])
    if not docs: docs = []
    backward_compatible = False

    if ids and not isinstance(ids, list):
        backward_compatible = True
        ids = [ids]

    if docs and not isinstance(docs, list):
        docs = [docs]

    final_result = []

    for id in ids:
        payload['id'] = id
        response = _record_index_similar(index, payload)
        final_result.append(response)

    for doc in docs:
        payload['doc'] = doc
        response = _record_index_similar(index, payload)
        final_result.append(response)

    if backward_compatible:
        return final_result[0]
    return final_result


def _record_index_similar(index, payload):
    query = prediction.create_prediction_query(payload)
    logger.debug(query)
    result = es.search(index=index, body=query)

    if not result.get('hits') or not result.get('hits').get('hits'):
        return {
            'total': 0,
            'result': [],
            'status_code': 200
        }

    result_records = result.get('hits').get('hits')
    response = {
        'total': result.get('hits').get('total').get('value'),
        "status_code": 200,
    }
    if 'aggregations' in result:
        response['aggs'] = score_based_aggregation(
            result_records, payload['aggs'])
    else:
        response['result'] = result_records
    return response


publish_options = {
    'publish_create': publish_create_index,
    'publish_modify': publish_modify_index,
    'publish_delete': publish_delete_index,
    'populate_index': publish_populate_index
}

index_options = {
    'index': record_index_data,
    'modify': record_index_update_data,
    'delete': record_index_delete,
    'search': record_index_search,
    'status': record_index_status,
    'similar': record_index_similar,
    'bulk_delete': record_index_bulk_delete
}

data_options = {
    'index': 'newData',
    'modify': 'newData',
    'delete': 'oldData',
    'search': 'payload',
    'bulk_delete': 'oldData'
}
