""" Copyright start
  Copyright (C) 2008 - 2022 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import requests
import time
from constants import APPLIANCE_PUBLIC_KEY, APPLIANCE_PRIVATE_KEY, MAX_RETRY_REQUESTS_COUNT
from cshmac.requests import HmacAuth
from util.logger import get_logger

logger = get_logger('__name__')


def maybe_json_or_raise(response):
    """
    Helper function for processing request responses

    Returns any json found in the response. Otherwise, it will extract the
    response as text, or, failing that, as bytes.

    :return: the response from the request
    :rtype: dict or str or bytes
    :raises: :class:`requests.HTTPError` if status code was 4xx or 5xx
    """
    if response and response.ok:
        try:
            return response.json(strict=False)
        except Exception:
            logger.warn(response.text or response.content)
            return {"hydra:totalItems": 0}
    else:
        return {"hydra:totalItems": 0}


def _make_request(method, url, body=None, auth=None, verify=False, allow_503=True, *args, **kwargs):
    headers = kwargs.get('headers', {})
    retry_count = 1
    response = None
    while retry_count < MAX_RETRY_REQUESTS_COUNT:
        try:
            response = requests.request(
                method,
                url,
                auth=HmacAuth(url, method, APPLIANCE_PUBLIC_KEY, APPLIANCE_PRIVATE_KEY),
                json=body,
                verify=verify,
                headers=headers,
                timeout=60)

            if response.status_code == 200:
                return maybe_json_or_raise(response)
            elif response.status_code in [401, 403, 404, 500]:
                logger.exception("Call for URL '%s' fails with status_code as '%s' and response text as '%s'" %(url, response.status_code, response.text))
                return {"hydra:totalItems": 0}
            elif response.status_code == '503' and not allow_503:
                retry_count += 1
            else:
                logger.exception("Call for URL '%s' fails with status_code as '%s' and response text as '%s'" %(url, response.status_code, response.text))
                return {"hydra:totalItems": 0}
        except requests.exceptions.SSLError:
            logger.exception("SSL Certificate Issue Reported")
            retry_count = MAX_RETRY_REQUESTS_COUNT
        except requests.exceptions.ConnectTimeout:
            retry_count += 1
            logger.exception("Connection TimeOut Occurred trying after 10 seconds")
        except requests.exceptions.ReadTimeout:
            retry_count += 1
            logger.exception("Connection ReadTimeOut Occurred trying after 10 seconds")
        except Exception as e:
            if(retry_count == 1):
                retry_count = (MAX_RETRY_REQUESTS_COUNT - 1)
                logger.exception("Generic Exception Occurred. Retrying to index data again")
            else:
                retry_count = MAX_RETRY_REQUESTS_COUNT
                logger.exception("Max retry counts exceeded. Exiting.")
        time.sleep(10)
    return maybe_json_or_raise(response)
